package application;

import guinterface.TimelineFrame;

import infrastructure.DeterministicNodeIdFactory;
import infrastructure.DirectDeliveryNotification;

import java.io.IOException;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import message.*;

import database.GossipAnalysisXMLDB;
import database.PartnerXMLDB;
import database.ProfileXMLDB;
import database.StatusXMLDB;
import database.TrafficAnalysisXMLDB;

import bogusmessage.ServerGetIDsMsg;
import bogusmessage.ServerGetStatusesMsg;
import bogusmessage.ServerGetUserMsg;
import bogusmessage.ServerRequestMsg;

import rice.p2p.commonapi.Application;
import rice.p2p.commonapi.DeliveryNotification;
import rice.p2p.commonapi.Endpoint;
import rice.p2p.commonapi.Id;
import rice.p2p.commonapi.Message;
import rice.p2p.commonapi.Node;
import rice.p2p.commonapi.NodeHandle;
import rice.p2p.commonapi.RouteMessage;
import rice.p2p.commonapi.rawserialization.InputBuffer;
import rice.p2p.commonapi.rawserialization.MessageDeserializer;
import rice.p2p.util.BloomFilter;
import rice.pastry.PastryNode;
import rice.pastry.PastryNodeFactory;

import toolbox.GossipLog;
import toolbox.HealthIndex;
import toolbox.Mailbox;
import toolbox.ObjectSerialize;
import toolbox.Partner;
import toolbox.PartnerMap;
import toolbox.TrafficAnalysisRecorder;
import toolbox.UTPair;

import winterwell.jtwitter.Twitter;
import winterwell.jtwitter.TwitterException;

public class Cuckoo implements Application {
	private static final int BULK = 20; //show BULK of tweets on the screen on time
	private static final int C = 3;
	
	private boolean GOSSIP_OR_NOT = false; //true: use gossip for broadcast
	private boolean UPLOAD_OR_NOT = false; //true: everything will upload to the twitter server
	
	protected PastryNodeFactory factory;
	protected Endpoint endpoint;
	protected Node node;
	protected String userId;
	protected Profile profile;

	//The XML-based database 
	private StatusXMLDB myStatusDB;
	private StatusXMLDB followeeStatusDB;
	private ProfileXMLDB myProfileDB;
	private ProfileXMLDB followeeProfileDB;
	private ProfileXMLDB followerProfileDB;
	private TrafficAnalysisXMLDB trafficAnalysisDB;
	private GossipAnalysisXMLDB gossipAnalysisDB;
	private PartnerXMLDB partnerXMLDB;
	
	//private FileOperator followPendings; //record the unsuccessful contacts for "follow" in the pending file
	//private XmlOperator  statusPendings; //record the unsuccessful contacts for "update status" in the pending file
	private String nodeHandleDir;

	private Mailbox mailbox = new Mailbox();
	
	//used for gossiping
	private PartnerMap partnerMap;
	private BloomFilter bloomFilter;
	private LinkedList<String> followeeIds;
	
	private Status latestStatus = null;
	
	private HealthIndex followeeHealthIndex;
	private HealthIndex followerHealthIndex;
	
	private TrafficAnalysisRecorder trafficRecorder;
	
	/////////////////////////////////////////////////
	/**
	 * GUI Components:
	 */
	private TimelineFrame timelineFrame;
	//////////////////////////////////////////////////
	
	//////////////////////////////////////////////////
	/**
	 * Twitter API the channel to the Twitter Server
	 */
	private Twitter twitter;
	//////////////////////////////////////////////////

	public static String nameExchange(String dirName) {
		return dirName.replaceAll("<", "").replace(">", "").replace(".", "");
	}
	
	public Cuckoo(PastryNodeFactory pnf, Node node, String uId, TimelineFrame tf, Twitter tw) {
		this.factory = pnf;
		this.endpoint = node.buildEndpoint(this, "cuckoo");
		this.node = node;
		this.userId = uId;
		
		this.timelineFrame = tf;
		this.twitter = tw;

		// Initialize the local XML database
		//String dir = nameExchange("xml/" + userId);
		//System.out.println("dir: " + dir);
		String  dir = "xml/" + userId;
		
		myStatusDB = new StatusXMLDB(dir, dir + "/mystatus.xml", "myStatus");
		followeeStatusDB = new StatusXMLDB(dir, dir + "/followeestatus.xml", "followeeStatus");
		myProfileDB = new ProfileXMLDB(dir, dir + "/myprofile.xml", "myprofile");
		followeeProfileDB = new ProfileXMLDB(dir, dir + "/followeeprofile.xml", "followeeprofile");
		followerProfileDB = new ProfileXMLDB(dir, dir + "/followerprofile.xml", "followerprofile");
		trafficAnalysisDB = new TrafficAnalysisXMLDB(dir, dir + "/trafficanalysis.xml", "trafficanalysis");
		gossipAnalysisDB  = new GossipAnalysisXMLDB(dir,  dir + "/gossipanalysis.xml",  "gossipanalysis");
		partnerXMLDB = new PartnerXMLDB(dir, dir + "/partner.xml", "partner");
		
		//record the target nodes that fails to follow
		//followPendings = new FileOperator(dir + "/fpending", followeeProfileDB);
		//records the target nodes that fails to get recent status from
		//statusPendings = new XmlOperator(dir, dir + "/spending.xml", "XmlOperator");
		nodeHandleDir = dir + "/followerHandleDir/";
		//create a buffer for the input message
		//out = ByteBuffer.wrap(node.getLocalNodeHandle().getId().toByteArray());
		//in = ByteBuffer.allocate(MSG_LENGTH);
		
		followeeHealthIndex = new HealthIndex(followeeProfileDB);
		followerHealthIndex = new HealthIndex(followerProfileDB);
		
		trafficRecorder = new TrafficAnalysisRecorder(trafficAnalysisDB);
		
		partnerMap = new PartnerMap(partnerXMLDB);
		bloomFilter = followeeProfileDB.buildBloomFilter();
		followeeIds = followeeProfileDB.getAllUserIds();
		
		endpoint.setDeserializer(new MessageDeserializer() {
		    
		      public Message deserialize(InputBuffer buf, short type, int priority, NodeHandle sender) throws IOException {
		        switch (type) {
		        	case ReqFollowMsg.TYPE:
		        		// just call the special deserialization constructor.
		        		return new ReqFollowMsg(buf, endpoint);
		        	case ReqStatusMsg.TYPE:
		        		return new ReqStatusMsg(buf, endpoint);
		        	case ReqUnfollowMsg.TYPE:
		        		return new ReqUnfollowMsg(buf, endpoint);
		        	case RplFollowErrMsg.TYPE:
		        		return new RplFollowErrMsg(buf, endpoint);
		        	case RplFollowMsg.TYPE:
		        		return new RplFollowMsg(buf, endpoint);
		        	case RplStatusErrMsg.TYPE:
		        		return new RplStatusErrMsg(buf, endpoint);
		        	case RplStatusMsg.TYPE:
		        		return new RplStatusMsg(buf, endpoint);
		        	case DirectUpdateStatusMsg.TYPE:
		        		return new DirectUpdateStatusMsg(buf, endpoint);
		        	case ReqNodeHandleMsg.TYPE:
		        		return new ReqNodeHandleMsg(buf, endpoint);
		        	case RplNodeHandleMsg.TYPE:
		        		return new RplNodeHandleMsg(buf, endpoint);
		        	case RplNodeHandleErrMsg.TYPE:
		        		return new RplNodeHandleErrMsg(buf, endpoint);
		        	case GossipUpdateStatusMsg.TYPE:
		        		return new GossipUpdateStatusMsg(buf, endpoint);
		        	case PartnerCancelMsg.TYPE:
		        		return new PartnerCancelMsg(buf, endpoint);
		        	case GossipFeedbackMsg.TYPE:
		        		return new GossipFeedbackMsg(buf, endpoint);
		        }
		        // maybe throw an IOException if you want
		        return null;
		      }    
		});
		
		//load the profile
		profile = myProfileDB.getProfile(userId);
		
		this.endpoint.register();
	}

	public static Id getIdFromString(String uId) {
		DeterministicNodeIdFactory nidFactory = new DeterministicNodeIdFactory(uId);
		return nidFactory.generateNodeId();
	}
	
	//When the followee is not online, the follower just "follow" it through Twitter Server
	//So every time, the peer rejoin the system, it checks the server for whether there are some new followees
	public void updateFollowerProfile() {
		new Thread() {
			public void run() {
				List<Long> idList = null;
				while(idList == null) {
					try{
						idList = twitter.getFollowerIDs();
						//---traffic log
						trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, ""), "s", "o", userId, "server", "n");
					} catch(Exception e) {
						int errcode = handleTwitterException(e);
						if(errcode == 403 || errcode == 404 || errcode == 500) {
							continue;
						}
					}
				}
				//---traffic log
				trafficRecorder.record(new ServerGetIDsMsg(idList, node.getId(), userId), "s", "i", "server", userId, "n");
		
				List<String> unhandlelist = new LinkedList<String>();
				if(idList == null) {
					return;
				}
				Iterator<Long> iterator = idList.iterator();
				while(iterator.hasNext()) {
					String tid = iterator.next().toString();
					if(followerProfileDB.exist(tid)) {
						break;
					} else {
						unhandlelist.add(tid);
					}
				}
		
				//Handle the unhandlelist
				Iterator<String> uIter = unhandlelist.iterator();
				while(uIter.hasNext()) {
					String uId = uIter.next();
					Twitter.User uprofile = null;
					while(uprofile == null) {
						try {
							uprofile = twitter.getUser(Long.valueOf(uId));
							//---traffic log
							trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, uId), "s", "o", userId, "server", "n");
						} catch(Exception e) {
							int errcode = handleTwitterException(e);
							if(errcode == 402 || errcode == 403 || errcode == 500) {
								continue;
							} else if(errcode == 604) {
								break;
							}
						}
					}
					//---traffic log
					trafficRecorder.record(new ServerGetUserMsg(uprofile, node.getId(), userId), "s", "i", "server", userId, "n");
					
					if(uprofile != null) {
						Profile p = new Profile(uprofile);
						if(timelineFrame != null) {timelineFrame.append2FollowerPanel(p.toString());}
						followerProfileDB.addProfile(p);
			
						Id nId = getIdFromString(uId);
						RplFollowMsg rplFollowMsg = new RplFollowMsg(node.getId(), nId, userId, uId, node.getLocalNodeHandle(), profile, generatePartners());
						
						reroute(nId, rplFollowMsg, null);
					}
				}
			}
		}.start();
	}
	
	public void updateFolloweeProfile() {
		new Thread() {
			public void run() {
				List<Long> idList = null;
				try{
					idList = twitter.getFriendIDs();
					//---traffic log
					trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, ""), "s", "o", userId, "server", "n");
				} catch(Exception e) {
					handleTwitterException(e);
				}
				//---traffic log
				trafficRecorder.record(new ServerGetIDsMsg(idList, node.getId(), userId), "s", "i", "server", userId, "n");
		
				List<String> unhandlelist = new LinkedList<String>();
				if(idList == null) {
					return;
				}
				Iterator<Long> iterator = idList.iterator();
				while(iterator.hasNext()) {
					String tid = iterator.next().toString();
					if(followeeProfileDB.exist(tid)) {
						break;
					} else {
						unhandlelist.add(tid);
					}
				}
				//Handle the unhandlelist
				Iterator<String> uIter = unhandlelist.iterator();
				while(uIter.hasNext()) {
					String uId = uIter.next();
					Twitter.User uprofile = null;
					while(uprofile == null) {
						try {
							uprofile = twitter.show(Long.valueOf(uId).longValue());
							//---traffic log
							trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, uId), "s", "o", userId, "server", "n");
						} catch(Exception e) {
							int errcode = handleTwitterException(e);
							if(errcode == 402 || errcode == 403 || errcode == 500) {
								continue;
							}  else if(errcode == 604) {
								break;
							}
						}
					}
					//---traffic log
					trafficRecorder.record(new ServerGetUserMsg(uprofile, node.getId(), userId), "s", "i", "server", userId, "n");
			
					if(uprofile != null) {
						Profile p = new Profile(uprofile);
						if(timelineFrame != null) {timelineFrame.append2FolloweePanel(p.toString());}
						followeeProfileDB.addProfile(p);
			
						Id nId = getIdFromString(uId);
						ReqFollowMsg reqFollowMsg = new ReqFollowMsg(node.getId(), nId, userId, uId, node.getLocalNodeHandle(), bloomFilter, profile);
						
						reroute(nId, reqFollowMsg, null);
					}
				}
			}
		}.start();
	}
	
	public void addMyProfile(Profile profile) {
		this.profile = profile;
		myProfileDB.addProfile(profile);
	}
	
	public void addMyProfile(String username, String location, String web, String bio) {
		profile = new Profile(userId, username, location, web, bio);
		myProfileDB.addProfile(profile);
	}

	public void follow(String uId) {//uId: userId (the followee)
		//If the user is already followed, return msg;
		if(followeeProfileDB.exist(uId) == true) {
			//timelineFrame.append2FolloweePanel("User " + uId + " is already followed!");
			System.out.println("User " + uId + " is already followed!");
			return;
		}
		
		if(profile == null) {
			System.out.println("profile is null!!!");
			System.exit(1);
		}
		
		Id nId = getIdFromString(uId);
		//Pack all the local information into the ReqFollowMsg
		ReqFollowMsg reqFollowMsg = new ReqFollowMsg(node.getId(), nId, userId, uId, node.getLocalNodeHandle(), bloomFilter, profile);
		reroute(nId, reqFollowMsg, null);

		if(timelineFrame != null) {timelineFrame.append2FolloweePanel("follow request is sent.\n");}
	}
	
	public void unfollow(String uId) {//uId: userId (the existing followee)
		if(followeeProfileDB.exist(uId) != true) {
			System.out.println("User " + uId + " is not followed by you!");
			return;
		}
		
		Id nId = getIdFromString(uId);
		
		//delete on Twitter
		String username = followeeProfileDB.getNameFromId(uId);
		twitter.stopFollowing(username);
		//---traffic log
		trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, username), "s", "o", userId, "server", "n");
		
		ReqUnfollowMsg reqUnfollowMsg = new ReqUnfollowMsg(node.getId(), nId, userId, uId, node.getLocalNodeHandle(), bloomFilter);
		
		//route the reqUnfollowMsg
		reroute(nId, reqUnfollowMsg, null);
		
		//delete on local DB
		followeeProfileDB.deleteProfile(uId);
	}
	
	public void writeStatus(String text) {
		if(text.compareTo("") == 0) {
			if(timelineFrame != null) {timelineFrame.showWarningDialog("The current Status is duplicate!");}
		} else if(latestStatus != null && text.compareTo(latestStatus.getText()) == 0) {
			if(timelineFrame != null) {timelineFrame.showWarningDialog("Duplicate status message!");}
			return;
		}
		
		Status s= null;
		if(UPLOAD_OR_NOT == true) {
			//First add on Twitter Server
			Twitter.Status sentStatus = null;
			while(sentStatus == null) {
				try {
					String first140text = getFirst140Text(text);
					sentStatus = twitter.updateStatus(first140text);
				} catch(Exception e) {
					int errcode = handleTwitterException(e);
					if(errcode == 403 || errcode == 404 ||errcode == 500) {
						continue;
					}
				}
			}
			LinkedList<Twitter.Status> stl = new LinkedList<Twitter.Status>();
			stl.add(sentStatus);
			trafficRecorder.record(new ServerGetStatusesMsg(stl, null, "server"), "s", "o", userId, "server", "n");
			trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, ""), "s", "i", "server", userId, "n");
		
			//System.out.println(calendar.getTime().toString());
			s = new Status(String.valueOf(sentStatus.id), text, profile.getUserName(), userId, sentStatus.createdAt.getTime());
		} else {
			s = new Status(String.valueOf(userId+new Date().getTime()), text, profile.getUserName(), userId, new Date().getTime());
		}
		
		if(timelineFrame != null) {timelineFrame.appendMyStatus(s.toString());}
		myStatusDB.addStatus(s);
		latestStatus = s;
		System.out.println(text);
		
		disseminate(s);
	}
	
	public void disseminate(Status s) {
		//LinkedList<String> followerIds = followerProfileDB.getAllUserIds();
		LinkedList<String> followerIds = followerHealthIndex.getHealthyIds();
		
		if(GOSSIP_OR_NOT == true) {
			gossipRandomSelection(followerIds);
		}
		
		if(GOSSIP_OR_NOT == true) {
			gossipAnalysisDB.addGossipLog(new GossipLog(new Date().getTime(), s.getStatusId(), 0, 0, followerHealthIndex.getNum()));
			disseminate(s, followerIds, true);
		} else {
			disseminate(s, followerIds, false);
		}
	}
	
	private void disseminate(Status s, List<String> ids, boolean gossip) {
		if(ids == null) {
			return;
		}
		
		Iterator<String> iterator = ids.iterator(); 
		while(iterator.hasNext()) {
			String followerId = iterator.next();
			//String path = nodeHandleDir+followerId;
			//NodeHandle nh = (NodeHandle) ObjectSerialize.file2object(path);
			NodeHandle nh = readFile2NodeHandle(followerId);
			
			if(nh == null) {
				System.out.println("nodeHandle is null, ask for nodeHandle!!!");
				s.setGossipOrNot(gossip);
				mailbox.add2mailbox(followerId, s);
				
				Id nId = getIdFromString(followerId);
				ReqNodeHandleMsg rnhmsg = new ReqNodeHandleMsg(node.getId(), nId, userId, followerId, node.getLocalNodeHandle(), bloomFilter);
				
				reroute(nId, rnhmsg, null);
				
			} else { //nh != null => means connect before
				/////////////////Test for exception//////////////////////
				//System.out.println("try to send to node [" + followerId + "]");
				//DirectDeliveryNotification ddn = new DirectDeliveryNotification(mailbox, s, followerId);
				//directRoute(nh, new UpdateStatusMsg(s, node.getId(), getIdFromString(followerId), userId, followerId), ddn);
				//////////////////////////////////////////////////////////
				
				int m = ((PastryNode) node).getLiveness((rice.pastry.NodeHandle)nh);
				if(m == 1) { //alive
					System.out.println("m == 1: " + "try to send to node [" + followerId + "]");
					if(gossip == false) {
						DirectUpdateStatusMsg dusm = new DirectUpdateStatusMsg(node.getId(), getIdFromString(followerId), userId, followerId, node.getLocalNodeHandle(), s);
						directRoute(nh, dusm, new DirectDeliveryNotification(mailbox, s, followerId, this, dusm), userId, followerId);
					} else { //GOSSIP_OR_NOT == true
						System.out.println("using gossip");
						GossipUpdateStatusMsg gusm = new GossipUpdateStatusMsg(node.getId(), getIdFromString(followerId), userId, followerId, node.getLocalNodeHandle(), bloomFilter, s);
						directRoute(nh, gusm, new DirectDeliveryNotification(mailbox, s, followerId, this, gusm), userId, followerId);
					}
				} else { //suspect or dead 
					System.out.println("m != 1: " + "try to send to node [" + followerId + "]");
					s.setGossipOrNot(gossip);
					mailbox.add2mailbox(followerId, s);
					//mailbox.printMailbox();
					//System.out.println("-------" + nh.getId().toString());
					Id nId = getIdFromString(followerId);
					//System.out.println("-----------followerId: " + followerId);
					//System.out.println("------------------nId: " + nId.toString());
					ReqNodeHandleMsg rnhmsg = new ReqNodeHandleMsg(node.getId(), nId, userId, followerId, node.getLocalNodeHandle(), bloomFilter);
					
					reroute(nId, rnhmsg, null);
				}
			}
		}
	}
	
	private void sendGossipAck(String oid, GossipFeedbackMsg gfm) {
		NodeHandle nh = readFile2NodeHandle(oid);
		Id nId = getIdFromString(oid);
		//-403 is used for mark the status is a feedback status
		Status dumpStatus = new Status(gfm.getStatusId(), null, null, "-403", 0, false);
		
		if(nh == null) {
			mailbox.add2mailbox(oid, dumpStatus);
			ReqNodeHandleMsg rnhmsg = new ReqNodeHandleMsg(node.getId(), nId, userId, oid, node.getLocalNodeHandle(), bloomFilter);
			reroute(nId, rnhmsg, null);
		} else { //nh != null => means connect before
			int m = ((PastryNode) node).getLiveness((rice.pastry.NodeHandle)nh);
			if(m == 1) { //alive
				directRoute(nh, gfm, new DirectDeliveryNotification(mailbox, dumpStatus, oid, this, gfm), userId, oid);
			} else { //suspect or dead
				mailbox.add2mailbox(oid, dumpStatus);
				ReqNodeHandleMsg rnhmsg = new ReqNodeHandleMsg(node.getId(), nId, userId, oid, node.getLocalNodeHandle(), bloomFilter);
				reroute(nId, rnhmsg, null);
			}
		}
	}
	
	public void gossipRandomSelection(LinkedList<String> followerIds) {
		//int gNum = (int) (Math.log( followerProfileDB.getNumofEntries() ) + C);
		int gNum = (int) (Math.log( followerHealthIndex.getNum() ) + C);
		gossipRandomSelection(followerIds, gNum);
	}
	
	private void gossipRandomSelection(LinkedList<String> followerIds, int num) {
		if(followerIds.size() <= num) {
			return;
		} else {
			int l = followerIds.size()-num;
			for(int i=0; i<l; i++) {
				int ran = (int) (Math.random()*(followerIds.size()));
				followerIds.remove(ran);
			}
		}
	}

	//For followees!
	public void addFollower(Profile fprofile) {
		followerProfileDB.addProfile(fprofile);
	}

//  //For testing 
//	public void deliver(Id id, Message message) {
//		System.out.println(this+" received "+message);
//	}
	
	public void updateFolloweeStatus() {
		LinkedList<UTPair> followees = followeeProfileDB.getAllUTPair();
		//If no followee
		if(followees == null || followees.size() == 0) {
			System.out.println("no followee!!");
			return;
		}
		
		//DEBUG
		//UTPair.printUTPairList(followees);
		
		UTPair pair;
		Iterator<UTPair> iterator = followees.iterator(); 
		while(iterator.hasNext()) {
			pair = iterator.next();
			updateFolloweeStatus(pair);
		}		
	}
	
	public void updateFolloweeStatus(String userId) {
		UTPair pair = followeeProfileDB.getUTPair(userId);
		if(pair != null) {
			updateFolloweeStatus(pair);
		}
	}
	
	private void updateFolloweeStatus(UTPair pair) { //userId and last time
		//pair.print();
		Id nId = getIdFromString(pair.userId);
		ReqStatusMsg reqStatusMsg = new ReqStatusMsg(node.getId(), nId, userId, pair.userId, node.getLocalNodeHandle(), bloomFilter, pair.time);
		
		reroute(nId, reqStatusMsg, null);	
		//statusPendings.add(pair.userId);
	}
	
	/**
	 * @Override: required by the "Application" interface
	 * Called when we receive a message.
	 */
	public void deliver(Id id, Message message) {
		if(!(message instanceof CuckooMsg)) {
			return;
		}
		
		CuckooMsg cm = (CuckooMsg) message;
		if( cm.getFromUserId() == userId ) {
			System.out.println("sent by myself...");
			return;
		}
		//Get the nodeHandler from the msg
		NodeHandle nh = cm.getNodeHandle();
		trafficRecorder.record(cm, "p", "i", cm.getFromUserId(), cm.getToUserId(), "n");
		
		if (message instanceof ReqFollowMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive ReqFollowMsg!");
			ReqFollowMsg msg = (ReqFollowMsg) message;

			if(msg.getToUserId().compareTo(userId) != 0) {
				System.out.println("The target of ReqFollowMsg is not online, my userId: " + userId + ", msg.userId: " + msg.getToUserId() + "***");
				
				//Return back the error msg to tell the offline condition!
				RplFollowErrMsg rplFollowErrMsg = new RplFollowErrMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), bloomFilter, msg.getToUserId());
				directRoute(nh, rplFollowErrMsg, null, rplFollowErrMsg.getFromUserId(), rplFollowErrMsg.getToUserId());
				return;
			}
			
			//Create Reply-Follow-Message that contains the newest profile; 
			RplFollowMsg profilemsg = new RplFollowMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), profile, generatePartners());
			addFollower(msg.getProfile()); //here init the BLOOD automatically
			writeNodeHandle2file(nh, msg.getProfile().getUserId());
			followerHealthIndex.healthIncrease(msg.getFromUserId());
			
			msg.getProfile().print();
			if(timelineFrame != null) {timelineFrame.append2FollowerPanel(msg.getProfile().toString());}
			
			//Send the Reply-Follow-Message to the follower as ACK;
			directRoute(nh, profilemsg, null, profilemsg.getFromUserId(), profilemsg.getToUserId());
			
		} else if(message instanceof RplFollowMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplFollowMsg!");
			RplFollowMsg msg = (RplFollowMsg) message;
			
			if(msg.getToUserId().compareTo(userId) != 0) {
				System.out.println("The target of RqlFollowMsg (" + msg.getToUserId() + ") is not online.");
				return; //do nothing!!
			}
			
			followeeProfileDB.addProfile(msg.getProfile()); //here init the BLOOD automatically
			followeeHealthIndex.addHealthIndex(msg.getProfile().getUserId());
			initPartnerMap(msg.getProfile().getUserId(), msg.getPartnerIds());
			
			if(timelineFrame != null) {timelineFrame.append2FolloweePanel(msg.getProfile().toString());}
			
			//follow on Twitter
			try {
				Twitter.User tuser = twitter.getUser(Long.valueOf(msg.getFromUserId()));
				//---traffic log
				trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, msg.getFromUserId()), "s", "o", userId, "server", "n");
				twitter.follow(tuser.screenName);
				//---traffic log
				trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, tuser.screenName), "s", "o", userId, "server", "n");
				
			} catch(Exception e) {
				handleTwitterException(e);
			}
			
		} else if(message instanceof ReqStatusMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive ReqStatusMsg!");
			ReqStatusMsg msg = (ReqStatusMsg) message;
			
			if(msg.getToUserId().compareTo(userId) != 0) {
				System.out.println("The target of ReqStatusMsg is offline, my userId: " + userId + ", msg.userId: " + msg.getToUserId() + "******");
				
				//Return back the msg that tells the offline condition!
				RplStatusErrMsg rplStatusErrMsg = new RplStatusErrMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), bloomFilter, msg.getToUserId());
				directRoute(nh, rplStatusErrMsg, null, rplStatusErrMsg.getFromUserId(), rplStatusErrMsg.getToUserId());
				return;
			}
			
			writeNodeHandle2file(nh, msg.getFromUserId());
			followerHealthIndex.healthIncrease(msg.getFromUserId());
			
			Status[] status = myStatusDB.getStatuses(msg.getTime());
			RplStatusMsg rplStatusMsg = new RplStatusMsg(node.getId(), msg.getTo(), userId, msg.getToUserId(), node.getLocalNodeHandle(), userId, status);
			
			//Send back the microblogs -- not care about whether that one can receive
			directRoute(nh, rplStatusMsg, null, rplStatusMsg.getFromUserId(), rplStatusMsg.getToUserId());
			
		} else if(message instanceof RplStatusMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplStatusMsg!");
			RplStatusMsg msg = (RplStatusMsg) message;
			
			String authorId = msg.getAuthorId();
			if(followeeProfileDB.exist(authorId) == false) {
				System.out.println("error!!! I don't need the msg!!!");
			}
			//The microblogs are in right track
			String authorName = followeeProfileDB.getNameFromId(authorId);
			
			String[] stid = msg.getRecentStids();
			String[] text = msg.getRecentTexts();
			long[] time = msg.getRecentTimes();
			
			if(stid == null) { return; }
			
			Status ns = null;
			for(int i=stid.length-1; i>=0; i--) {
				ns = new Status(stid[i], text[i], authorName, authorId, time[i]);
				//For GUI :-( ugly!
				if(timelineFrame != null) {timelineFrame.appendFolloweeStatus(ns.toString());}
				followeeStatusDB.addStatus(ns);
			}
			followeeProfileDB.updateTime(ns.getAuthorId(), String.valueOf(ns.getTime()));
			followeeHealthIndex.healthIncrease(authorId);
			
		} else if(message instanceof DirectUpdateStatusMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive DirectUpdateStatusMsg!");
			DirectUpdateStatusMsg msg = (DirectUpdateStatusMsg) message;
			
			System.out.println("DirectUpdateStatusMsg.getFromUserId: --" + msg.getFromUserId());
			//System.out.println("DirectUpdateStatusMsg.getFrom: " + msg.getFrom().toString());
			System.out.println("DirectUpdateStatusMsg.getToUserId: ----" + msg.getToUserId());
			//System.out.println("DirectUpdateStatusMsg.getTo: " + msg.getTo());
			System.out.println("DirectUpdateStatusMsg.text: -----------" + msg.getText());
			
			String authorId = msg.getAuthorId();
			
			if(followeeProfileDB.exist(authorId) == false) {
				System.out.println("error!!! I don't need the msg!!!");
				ReqUnfollowMsg reqUnfollowMsg = new ReqUnfollowMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), bloomFilter);
				reroute(msg.getFrom(), reqUnfollowMsg, null);		
				return;
			}
			
			String authorName = followeeProfileDB.getNameFromId(authorId);
			//debugging
			System.out.println(this.userId + " receive " + msg.getText() + " from " + authorName);
			
			//For GUI :-( ugly!
			Status ns = new Status(msg.getStatusId(), msg.getText(), authorName, authorId, msg.getTime());
			
			if(timelineFrame != null) {timelineFrame.appendFolloweeStatus(ns.toString());}
			followeeStatusDB.addStatus(ns);
			followeeProfileDB.updateTime(ns.getAuthorId(), String.valueOf(ns.getTime()));
			followeeHealthIndex.healthIncrease(authorId);			
			
		} else if(message instanceof GossipUpdateStatusMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive GossipUpdateStatusMsg!");
			GossipUpdateStatusMsg msg = (GossipUpdateStatusMsg) message;
			
			System.out.println("GossipUpdateStatusMsg.getFromUserId: --" + msg.getFromUserId());
			System.out.println("GossipUpdateStatusMsg.getToUserId: ----" + msg.getToUserId());
			System.out.println("GossipUpdateStatusMsg.text: -----------" + msg.getText());
			
			String authorId = msg.getAuthorId();
			
			if(followeeProfileDB.exist(authorId) == false) {
				System.out.println("error!!! I don't need the msg!!!");
				//cancel the partnership
				PartnerCancelMsg partnerCancelMsg = new PartnerCancelMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), authorId);
				directRoute(nh, partnerCancelMsg, null, userId, msg.getFromUserId());
				return;
			}
			
			String authorName = followeeProfileDB.getNameFromId(authorId);
			Status ns = new Status(msg.getStatusId(), msg.getText(), authorName, authorId, msg.getTime());
			if(followeeStatusDB.exist(ns) == false) {
				
				/*first show on the interface and add to local db*/
				if(timelineFrame != null) {timelineFrame.appendFolloweeStatus(ns.toString());}
				followeeStatusDB.addStatus(ns);
				followeeProfileDB.updateTime(ns.getAuthorId(), String.valueOf(ns.getTime()));
				followeeHealthIndex.healthIncrease(authorId);
				
				/*send back the recv msg to the original user :-(*/
				sendGossipAck(authorId, new GossipFeedbackMsg(node.getId(), getIdFromString(authorId), userId, authorId, node.getLocalNodeHandle(), msg.getStatusId()));
				
				/*gossip now, ahahahahaha!*/
				List<String> partners = partnerMap.getPartners(msg.getAuthorId());
				
				//System.out.println("///////////////////////////////////////////////////////////");
				//printStringList(partners);
				
				disseminate(ns, partners, true);
			} else {
				System.out.println("Redundant message: from [" + msg.getFromUserId() + "] to [" + msg.getToUserId() + "], text: " + msg.getText());
				trafficRecorder.record(cm, "p", "i", cm.getFromUserId(), cm.getToUserId(), "r");
			}
		} else if(message instanceof RplFollowErrMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplFollowErrMsg!");
			RplFollowErrMsg msg = (RplFollowErrMsg) message;
			
			if(msg.getToUserId().compareTo(userId) != 0) {
				System.out.println("The target of RqlFollowErrMsg (" + msg.getToUserId() + ") is not online.");
				return;
			}
			
			//The follow target is offline, just request to server
			Twitter.User tuser = null;
			try {
				tuser = twitter.getUser(Long.valueOf(msg.getUserId()));
				//---traffic log
				trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, msg.getUserId()), "s", "o", userId, "server", "n");
				twitter.follow(tuser.screenName);
				//---traffic log
				trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, tuser.screenName), "s", "o", userId, "server", "n");
			} catch(Exception e) {
				handleTwitterException(e);
			}
			Profile p = new Profile(tuser);
			
			followeeProfileDB.addProfile(p);
			followeeHealthIndex.addHealthIndex(p.getUserId());
			followeeHealthIndex.healthDecrease(p.getUserId());
			if(timelineFrame != null) {timelineFrame.append2FolloweePanel(p.toString());}
			
		} else if(message instanceof RplStatusErrMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplStatusErrMsg!");
			//The request target is offline, re-evaluate the H-index
			RplStatusErrMsg msg = (RplStatusErrMsg) message;
			
			if(msg.getToUserId().compareTo(userId) == 0) {
				followeeHealthIndex.healthDecrease(msg.getUserId());
				fetchTweetFromServer(msg.getUserId());
			}
		} else if(message instanceof ReqUnfollowMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive ReqUnfollowMsg!");
			ReqUnfollowMsg msg = (ReqUnfollowMsg) message;
			
			if(msg.getToUserId().compareTo(userId) != 0) {
				return; //do nothing, just discard.
			}
			followerProfileDB.deleteProfile(msg.getFromUserId());
			followerHealthIndex.deleteHealthIndex(msg.getFromUserId());
			
		} else if(message instanceof ReqNodeHandleMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive ReqNodeHandleMsg!");
			ReqNodeHandleMsg msg = (ReqNodeHandleMsg) message;
			
			System.out.println("ReqNodeHandleMsg.getFromUserId: " + msg.getFromUserId());
			//System.out.println("ReqNodeHandleMsg.getFrom: " + msg.getFrom().toString());
			System.out.println("ReqNodeHandleMsg.getToUserId: " + msg.getToUserId());
			//System.out.println("ReqNodeHandleMsg.getTo: " + msg.getTo());
			
			if(msg.getToUserId().compareTo(userId) != 0) {
				System.out.println("target offline!");
				//System.out.println("msg.getToUserId() " + msg.getToUserId() +", userId: " + userId);
				directRoute(nh, new RplNodeHandleErrMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), bloomFilter, msg.getToUserId()), null, userId, msg.getFromUserId());
				return;
			}
			directRoute(nh, new RplNodeHandleMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle()), null, userId, msg.getFromUserId());
			
		} else if(message instanceof RplNodeHandleMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplNodeHandleMsg!");
			RplNodeHandleMsg msg = (RplNodeHandleMsg) message;
			
			System.out.println("RplNodeHandleMsg.getFromUserId: " + msg.getFromUserId());
			//System.out.println("RplNodeHandleMsg.getFrom: " + msg.getFrom().toString());
			System.out.println("RplNodeHandleMsg.getToUserId: " + msg.getToUserId());
			//System.out.println("RplNodeHandleMsg.getTo: " + msg.getTo());
			
			if(msg.getToUserId().compareTo(userId) == 0) {
				writeNodeHandle2file(nh, msg.getFromUserId());
				
				//clear the log send
				LinkedList<Status> alist = mailbox.getStatuses(msg.getFromUserId());
				if(alist != null) {
					while(!alist.isEmpty()) {
						Status s = alist.removeFirst();
						//System.out.println(s.toString());
						if(s.getAuthorId().compareTo("-403") == 0) {
							GossipFeedbackMsg gfm = new GossipFeedbackMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), s.getStatusId());
							directRoute(nh, gfm, new DirectDeliveryNotification(mailbox, s, msg.getFromUserId(), this, gfm), userId, msg.getFromUserId());
							return;
						} 
						
						if(s.getGossipOrNot() == false) {
							DirectUpdateStatusMsg dusm = new DirectUpdateStatusMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), s);
							directRoute(nh, dusm, new DirectDeliveryNotification(mailbox, s, msg.getFromUserId(), this, dusm), userId, msg.getFromUserId());
						} else {
							GossipUpdateStatusMsg gusm = new GossipUpdateStatusMsg(node.getId(), msg.getFrom(), userId, msg.getFromUserId(), node.getLocalNodeHandle(), bloomFilter, s);
							directRoute(nh, gusm, new DirectDeliveryNotification(mailbox, s, msg.getFromUserId(), this, gusm), userId, msg.getFromUserId());
						}
					}
				} else {
					System.out.println(msg.getFromUserId() + "in mailbox is null");
				}
			}
		} else if(message instanceof RplNodeHandleErrMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive RplNodeHandleErrMsg!");
			RplNodeHandleErrMsg msg = (RplNodeHandleErrMsg) message;
			
			System.out.println("RplNodeHandleErrMsg.getFromUserId: " + msg.getFromUserId());
			//System.out.println("RplNodeHandleErrMsg.getFrom: " + msg.getFrom().toString());
			System.out.println("RplNodeHandleErrMsg.getToUserId: " + msg.getToUserId());
			//System.out.println("RplNodeHandleErrMsg.getTo: " + msg.getTo());
			
			/*if(msg.getToUserId().compareTo(userId) == 0) {
				String userid = msg.getOriginId();
				//decrease the H-index
				followerHealthIndex.hDecrease(userid);
			}*/
		} else if(message instanceof PartnerCancelMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive PartnerCancelMsg!");
			PartnerCancelMsg msg = (PartnerCancelMsg) message;
			
			if(msg.getToUserId().compareTo(userId) == 0) {
				partnerMap.deletePartner(msg.getCommonFolloweeId(), msg.getFromUserId());
			}
		} else if(message instanceof GossipFeedbackMsg) {
			System.out.println("Node " + profile.getUserName() + " [" + userId + "] " + "receive GossipFeedbackMsg!");
			GossipFeedbackMsg msg = (GossipFeedbackMsg) message;
			
			if(msg.getToUserId().compareTo(userId) == 0) { // else do nothing! (just discard)
				increaseGossipRecv(msg.getStatusId());
			}
		}
	}

	/**
	 * Called when a message is traveling along your path. 
	 */
	public boolean forward(RouteMessage message) {
		try {
			Message mesg = message.getMessage(endpoint.getDeserializer());
			
			if(mesg instanceof CuckooPiggybackMsg) {
				CuckooPiggybackMsg msg = (CuckooPiggybackMsg) mesg;
				
				if(msg.getFromUserId().compareTo(userId) == 0) { //message from myself or to myself
					return true;
				}
				
				if(msg.getToUserId().compareTo(userId) != 0) {
					trafficRecorder.record(msg, "f", "i", msg.getFromUserId(), msg.toString(), "n");
				}
				System.out.println("----Node " + profile.getUserName() + " [" + userId + "] forward----: " + msg.toString());
				decidePartner(msg);
			}
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
		return true;
	}

	/**
	 * Called when you hear about a new neighbor. Don't worry about this method
	 * for now.
	 */
	public void update(NodeHandle arg0, boolean arg1) {
	}

	/**
	 * Called to directly send a message to the nh
	 * The receiver receive the msg by "deliver" function
	 */
	public void directRoute(NodeHandle nh, Message msg, DeliveryNotification dn, String fid, String tid) {
		//System.out.println(this+" sending direct to "+nh);
		endpoint.route(null, msg, nh, dn);
		//---traffic log
		trafficRecorder.record(msg, "p", "o", fid, tid, "n");
	}
	
	public void reroute(Id destId, CuckooMsg msg, DeliveryNotification dn) {
		endpoint.route(destId, msg, null, dn);
		trafficRecorder.record(msg, "p", "o", msg.getFromUserId(), msg.getToUserId(), "n");
	}
	
	//return the NodeHandle of a specific user with uid
	public NodeHandle readFile2NodeHandle(String uid) {
		return (NodeHandle) ObjectSerialize.file2object(nodeHandleDir+uid);
	}
	
	//write NodeHandle to file (serialization)
	public void writeNodeHandle2file(NodeHandle nh, String uid) {
		ObjectSerialize.object2file(nh, nodeHandleDir+uid);
	}
	
	public void deleteNodeHandle(String id) {
		
	}
		
	public String toString() {
		return "Cuckoo (nodeId: " + endpoint.getId() + ", userId: " + userId + ")";
	}
	
	public void quit() {
		System.out.println("Bye, Bye!!");
		//write back the healthIndex
		followerHealthIndex.writeBack();
		followeeHealthIndex.writeBack();
		System.exit(1);
	}
	
	public void decidePartner(CuckooPiggybackMsg cpmsg) {
		decidePartner(cpmsg.getFromUserId(), cpmsg.getBloomFilter(), cpmsg.getNodeHandle());
	}
	
	private void decidePartner(String pid, BloomFilter bf, NodeHandle nh) {
		Iterator<String> iterator = followeeIds.iterator();
		Partner p = null;
		LinkedList<String> ll = new LinkedList<String>();
		while(iterator.hasNext()) {
			String fid = iterator.next();
			if(bf.check(fid.getBytes()) == true) {
				System.out.println("common interest: " + fid);
				if(p == null) {
					p = new Partner(fid, pid, fid, 1, HealthIndex.INIT_BLOOD);
				} else {
					p.addCommonId(fid);
					p.addSimilarity();
				}
				ll.add(fid);
			}
		}
		iterator = ll.iterator();
		while(iterator.hasNext()) {
			partnerMap.runtimeAdd(iterator.next(), p);
		}
		
		//Debugging
		printPartnerMap();
	}
	
	public String generatePartners() {
		LinkedList<String> followerIds = followerHealthIndex.getHealthyIds();
		gossipRandomSelection(followerIds, PartnerMap.NUM_OF_PARTNERS);
		
		String ids = "";
		Iterator<String> iterator = followerIds.iterator();
		while(iterator.hasNext()) {
			ids = ids + iterator.next() + "#";
		}
		System.out.println("generatePartners ------------------ " + ids);
		return ids;
	}
	
	public void initPartnerMap(String fid, String pids) {
		String[] split = pids.split("#");
		for(int i=0; i<split.length; i++) {
			partnerMap.initAdd(fid, new Partner(fid, split[i], fid, 1, HealthIndex.INCR_BLOOD));		
		}
		partnerMap.writeBack();
	}
	
	public void increaseGossipSent(String stId) {
		gossipAnalysisDB.increaseSentMsg(stId);
	}
	
	public void increaseGossipRecv(String stId) {
		gossipAnalysisDB.increaseRecvMsg(stId);
	}
	
	public void increaseFollowerHealth(String fid) {
		followerHealthIndex.addHealthIndex(fid);
	}
	
	public void decreaseFollowerHealth(String fid) {
		followerHealthIndex.deleteHealthIndex(fid);
	}
	
	public void increasePartnerHealth(String pid, String uid) {
		partnerMap.increaseHealth(pid, uid);
	}
	
	public void decreasePartnerHealth(String pid, String uid) {
		partnerMap.decreaseHealth(pid, uid);
	}
	
	//////////////Twitter Staff /////////////////
	
	public int handleTwitterException(Exception e) {
		if(e instanceof TwitterException.E401) {
			if(timelineFrame != null) {timelineFrame.showWarningDialog("Authorication is wrong!");}
			return 401;
		} else if (e instanceof TwitterException.E403) {
			//do nothing
			return 403;
		} else if (e instanceof TwitterException.E404) {
			if(timelineFrame != null) {timelineFrame.showWarningDialog("Resource does not exist on Twitter!");}
			return 404;
		} else if (e instanceof TwitterException.E50X) {
			if(timelineFrame != null) {timelineFrame.showWarningDialog("Twitter has some problems!");}
			return 500;
		} else if (e instanceof TwitterException.Repetition) {
			if(timelineFrame != null) {timelineFrame.showWarningDialog("Duplicate msg!!");}
			return 601;
		} else if (e instanceof TwitterException.RateLimit) {
			if(timelineFrame != null) {timelineFrame.showWarningDialog("Ooops, you reach the rate limit!");}
			return 602;
		} else if (e instanceof TwitterException.Timeout) {
			System.out.println("probably caused by Twitter being overloaded");
			return 603;
		} else if (e instanceof NullPointerException) {
			System.out.println("The Id is problematic");
			return 604;
		}
		return -1;
	}
	
	public void fetchTweetFromServer() {
		System.out.println("fetch all followees' tweets from server");
		Date lastDate = new Date(followeeStatusDB.getLastTime());
		//System.out.println(lastDate.toString());
		List<Twitter.Status> stl = null;
		while(stl == null) {
			try {
				twitter.setSinceDate(lastDate);
				stl = twitter.getFriendsTimeline();
				//---traffic log
				trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, ""), "s", "o", userId, "server", "n");
				trafficRecorder.record(new ServerGetStatusesMsg(stl, node.getId(), userId), "s", "i", "server", userId, "n");		
			} catch(Exception e) {
				int errcode = handleTwitterException(e);
				if(errcode == 403 || errcode == 500) {
					continue;
				} else {
					break;
				}
			}
		}
		Iterator<Twitter.Status> iterator = stl.iterator();
		while(iterator.hasNext()) {
			Twitter.Status st = iterator.next();
			System.out.println(st.getUser().screenName + " " + st.text);
			if(userId.compareTo(String.valueOf(st.getUser().id)) != 0) {
				Status s = new Status(st);
				if(timelineFrame != null) {timelineFrame.appendFolloweeStatus(s.toString()); }
				followeeStatusDB.addStatus(s);
				followeeProfileDB.updateTime(st);
			}
		}
	}
	
	public void fetchTweetFromServer(String userid) {
		String username = followeeProfileDB.getNameFromId(userid);
		System.out.println("fetch followee " + username + "'s tweets from server");
		
		//Fetch from the Twitter server
		if(username == null) {
			System.out.println("This guy is not my followee!!");
			return;
		}
		
		Date lastDate = new Date(followeeProfileDB.getLastTime(userid));
		List<Twitter.Status> stl = null;
		while(stl == null) {
			try {
				twitter.setSinceDate(lastDate);
				stl = twitter.getUserTimeline(username);
				//---traffic log
				trafficRecorder.record(new ServerRequestMsg(node.getId(), userId, ""), "s", "o", userId, "server", "n");
				trafficRecorder.record(new ServerGetStatusesMsg(stl, node.getId(), userId), "s", "i", userId, "server", "n");
			} catch(Exception e) {
				int errcode = handleTwitterException(e);
				if(errcode == 403 || errcode == 500) {
					continue;
				} else {
					break;
				}
			}
		}
		if(stl == null) {return;}
		Iterator<Twitter.Status> iterator = stl.iterator();
		while(iterator.hasNext()) {
			Twitter.Status st = iterator.next();
			System.out.println(st.getUser().screenName + " " + st.text);
			if(userId.compareTo(String.valueOf(st.getUser().id)) != 0) {
				Status s = new Status(st);
				//timelineFrame.appendFolloweeStatus(s.toString());
				followeeStatusDB.addStatus(s);
				followeeProfileDB.updateTime(st);
			}
		}
	}
	
	private String getFirst140Text(String text) {
		if(text.length() > 140) {
			return text.substring(0, 139);
		} else {
			return text;
		}
	}
	
	//////////////Getters/////////////////
	
	public Node getNode() {
		return node;
	}
	
	public String getUserId() {
		return userId;
	}
	
	public Id getNodeId() {
		return node.getId();
	}
	
	public NodeHandle getNodeHandle() {
		return node.getLocalNodeHandle();
	}
	
	public TrafficAnalysisXMLDB getTrafficAnalysisXMLDB() {
		return trafficAnalysisDB;
	}
	
	public GossipAnalysisXMLDB getGossipAnalysisXMLDB() {
		return gossipAnalysisDB;
	}
	
	public Status[] getFolloweeStatus() {
		return followeeStatusDB.getRecentStatus(BULK);
	}
	
	public LinkedList<Profile> getFolloweeProfile() {
		return followeeProfileDB.getAllProfile();
	}
	
	public LinkedList<Profile> getFollowerProfile() {
		return followerProfileDB.getAllProfile();
	}
	
	public Status[] getMyStatus() {
		Status[] ss = myStatusDB.getRecentStatus(BULK);
		if(ss == null || ss.length == 0) {
			return null;
		}
		latestStatus = ss[0];
		//System.out.print(latestStatus.toString());
		return ss;
	}
	
	public BloomFilter getBloomFilter() {
		return bloomFilter;
	}
	
	public boolean getGossipOrNot() {
		return GOSSIP_OR_NOT;
	}
	
	public boolean getUploadOrNot() {
		return UPLOAD_OR_NOT;
	}
	
	public Profile getProfile() {
		return profile;
	}
	
	//////////////Printers/////////////////
	
	public void printProfile() {
		myProfileDB.xmlParsePrint();
	}
	
	public void printMyStatus() {
		StatusXMLDB.printStatuses(myStatusDB.getRecentStatus(BULK));
	}
	
	public void printFolloweeStatus() {
		StatusXMLDB.printStatuses(followeeStatusDB.getRecentStatus(BULK));
	}
	
	public void printFollowers() {
		followerProfileDB.xmlParsePrint();
	}
	
	public void printFollowees() {
		followeeProfileDB.xmlParsePrint();
	}
	
	public void printGossipOrNot() {
		System.out.println(GOSSIP_OR_NOT);
	}
	
	public static void printStringList(List<String> l) {
		Iterator<String> iterator = l.iterator();
		while(iterator.hasNext()) {
			System.out.print(iterator.next() + "  ");
		}
		System.out.println();
	}
	
	public void printFolloweeHealthIndex() {
		System.out.println("----------Followee Health Index----------");
		followeeHealthIndex.printHeathIndice();
	}
	
	public void printFollowerHealthIndex() {
		System.out.println("----------Follower Health Index----------");
		followerHealthIndex.printHeathIndice();
	}
	
	public void printPartnerMap() {
		System.out.println("---------------Partner Map---------------");
		partnerMap.printPartnerMap();
	}
	
	////////Setters///////////////////////
	
	public void setGossipOrNot(boolean g) {
		GOSSIP_OR_NOT = g;
	}
	
	public void setUploadOrNot(boolean u) {
		UPLOAD_OR_NOT = u;
	}
}
